/*
 * Copyright (c) 2004, 2005, 2006, 2007 Luiz Otavio O Souza <loos.br@gmail.com>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

static const char rcsid[] = "$Id: check-cmd.c,v 1.76 2007/07/28 18:08:01 luiz Exp $";

#include <errno.h>
#include <stdio.h>
#include <string.h>

#include "io.h"
#include "ns.h"
#include "sb.h"
#include "msg.h"
#include "user.h"
#include "client.h"
#include "server.h"
#include "return.h"
#include "command.h"
#include "contacts.h"
#include "protocol.h"
#include "msn-proxy.h"

typedef struct {
    char	*cmd;
    int		args;
    int		(*func)(user_ *user, command *cmd, int args);
} CMD;

typedef struct {
    char	*cmd;
    int		args;
    int		(*func)(sb_ *sb, command *cmd, int args);
} SBCMD;

int check_ok(user_ *user, command *cmd, int args) { return(ROK); }

int _check_args(command *cmd, int args) {

    if (args != cmd->args_len) {
        io_printf(1, "numero de args: [%d] != [%d]\n", args, cmd->args_len);
	print_command(cmd);
        return(RFAIL);
    }
    return(ROK);
}

int check_args(user_ *user, command *cmd, int args) {
    return(_check_args(cmd, args));
}

int check_sb_args(sb_ *sb, command *cmd, int args) {
    return(_check_args(cmd, args));
}

CMD	SERVER_PRE_CMDS[] = {
		{ "VER", 0, server_pre_auth_ver },
		{ NULL, 0, NULL }
	};

CMD	SERVER_MSNP8_PRE_CMDS[] = {
		{ "CVR", 6, check_args },
		{ "XFR", 0, xfr_proxy },
		{ "USR", 0, server_msnp8_pre_usr },
		{ NULL, 0, NULL }
	};

CMD	SERVER_MSNP8_POST_CMDS[] = {
		{ "XFR", 0, xfr_ns_post_proxy },
		{ NULL, 0, NULL }
	};

CMD	SERVER_MSNP8_CMDS[] = {
		{ "MSG", 3, check_args },
		{ "SYN", 0, msnp8_server_syn_rewrite },
		{ "GTC", 1, check_args },
		{ "BLP", 0, check_ok },
		{ "PRP", 2, check_args },
		{ "LSG", 0, msnp8_save_contactgroup },
		{ "LST", 0, msnp8_save_contact },
		{ "BPR", 0, check_ok },
		{ "CHG", 0, check_ok },
		{ "ILN", 0, msn_contact_initial_state },
		{ "CHL", 2, check_args },
		{ "FLN", 0, msn_contact_logoff },
		{ "NLN", 0, msn_contact_change },
		{ "QRY", 1, check_args },
		{ "QNG", 1, check_args },
		{ "XFR", 0, xfr_proxy },
		{ "RNG", 0, rng_sb_proxy },
		{ "URL", 0, msn_url },
		{ "ADD", 0, check_ok }, /* XXX - add users */
		{ "REA", 0, check_ok }, /* XXX - rename users */
		{ "REM", 0, check_ok }, /* XXX - remove users */
		{ "NOT", 1, check_args },
		{ "OUT", 1, check_args },
		{ NULL, 0, NULL }
	};

CMD	SERVER_MSNP12_PRE_CMDS[] = {
		{ "CVR", 6, check_args },
		{ "XFR", 0, xfr_proxy },
		{ "USR", 0, server_msnp12_pre_usr },
		{ NULL, 0, NULL }
	};

CMD	SERVER_MSNP12_POST_CMDS[] = {
		{ "XFR", 0, xfr_ns_post_proxy },
		{ NULL, 0, NULL }
	};

CMD	SERVER_MSNP12_CMDS[] = {
		{ "SBS", 2, check_args },
		{ "MSG", 3, check_args },
		{ "SYN", 0, check_ok },
		{ "GCF", 0, check_ok },
		{ "CHG", 0, check_ok },
		{ "ILN", 0, msn_contact_initial_state },
		{ "UUX", 2, check_args },
		{ "UBX", 0, msnp12_ubx },
		{ "URL", 0, msn_url },
		{ "CHL", 2, check_args },
		{ "FLN", 0, msn_contact_logoff },
		{ "QRY", 1, check_args },
		{ "NLN", 0, msn_contact_change },
		{ "QNG", 1, check_args },
		{ "XFR", 0, xfr_proxy },
		{ "GTC", 1, check_args },
		{ "BLP", 1, check_args },
		{ "PRP", 0, msnp12_update_dn },
		{ "LSG", 0, msnp12_save_contactgroup },
		{ "LST", 0, msnp12_save_contact },
		{ "BPR", 0, check_ok },
		{ "SBP", 4, check_args },
		{ "RNG", 0, rng_sb_proxy },
		{ "ADC", 0, check_ok }, /* XXX */
		{ "ADG", 0, check_ok }, /* XXX */
		{ "REM", 0, check_ok }, /* XXX */
		{ "NOT", 1, check_args },
		{ "OUT", 1, check_args },
		{ NULL, 0, NULL }
	};

CMD	SERVER_MSNP13_PRE_CMDS[] = {
		{ "CVR", 6, check_args },
		{ "XFR", 0, msnp13_ns_xfr_proxy },
		{ "GCF", 2, check_args },
		{ "USR", 0, server_msnp12_pre_usr },
		{ NULL, 0, NULL }
	};

CMD	SERVER_MSNP13_POST_CMDS[] = {
		{ "XFR", 0, xfr_ns_post_proxy },
		{ NULL, 0, NULL }
	};

CMD	SERVER_MSNP13_CMDS[] = {
		{ "RML", 2, check_args },
		{ "FQY", 2, check_args },
		{ "SBS", 2, check_args },
		{ "MSG", 3, check_args },
		{ "BLP", 2, check_args },
		{ "ADL", 2, check_args },
		{ "CHG", 0, check_ok },
		{ "ILN", 0, msnp13_contact_initial_state },
		{ "UUX", 2, check_args },
		{ "UUN", 2, check_args },
		{ "UBN", 3, check_args },
		{ "UBX", 0, msnp12_ubx },
		{ "URL", 0, msn_url },
		{ "CHL", 2, check_args },
		{ "FLN", 0, msn_contact_logoff },
		{ "QRY", 1, check_args },
		{ "NLN", 0, msnp13_contact_change },
		{ "QNG", 1, check_args },
		{ "XFR", 0, xfr_sb_proxy },
		{ "PRP", 0, msnp12_update_dn },
		{ "RNG", 0, rng_sb_proxy },
		{ "OUT", 1, check_args },
		{ NULL, 0, NULL }
	};

CMD	CLIENT_PRE_CMDS[] = {
		{ "VER", 0, client_pre_auth_ver },
		{ NULL, 0, NULL }
	};

CMD	CLIENT_MSNP8_PRE_CMDS[] = {
		{ "CVR", 9, check_args },
		{ "OUT", 0, check_ok },
		{ "USR", 0, client_msn_pre_usr },
		{ NULL, 0, NULL }
	};

CMD	CLIENT_MSNP8_POST_CMDS[] = {
		{ NULL, 0, NULL }
	};

CMD	CLIENT_MSNP8_CMDS[] = {
		{ "SYN", 0, msnp8_client_syn_rewrite },
		{ "CHG", 0, msn_update_status },
		{ "QRY", 3, check_args },
		{ "PNG", 0, check_ok },
		{ "XFR", 2, check_args },
		{ "BLP", 2, check_args },
		{ "ADD", 0, check_ok }, /* XXX - add lists */
		{ "REA", 0, check_ok }, /* XXX - add lists */
		{ "REM", 0, check_ok }, /* XXX - add lists */
		{ "URL", 0, msn_url },
		{ "OUT", 0, check_ok },
		{ NULL, 0, NULL }
	};

CMD	CLIENT_MSNP12_PRE_CMDS[] = {
		{ "CVR", 9, check_args },
		{ "USR", 0, client_msn_pre_usr },
		{ "OUT", 0, check_ok },
		{ NULL, 0, NULL }
	};

CMD	CLIENT_MSNP12_POST_CMDS[] = {
		{ NULL, 0, NULL }
	};

CMD	CLIENT_MSNP12_CMDS[] = {
		{ "GCF", 2, check_args },
		{ "SYN", 0, msnp12_client_syn_rewrite },
		{ "CHG", 0, msn_update_status },
		{ "UUX", 2, check_args },
		{ "URL", 0, msn_url },
		{ "PNG", 0, check_ok },
		{ "QRY", 3, check_args },
		{ "XFR", 2, check_args },
		{ "PRP", 3, check_args },
		{ "SBP", 4, check_args },
		{ "OUT", 0, check_ok },
		{ "CHL", 1, check_args },
		{ "ADC", 0, check_ok }, /* XXX */
		{ "ADG", 0, check_ok }, /* XXX */
		{ "REA", 0, check_ok }, /* XXX - add lists */
		{ "REM", 0, check_ok }, /* XXX - add lists */
		{ NULL, 0, NULL }
	};

CMD	CLIENT_MSNP13_PRE_CMDS[] = {
		{ "OUT", 0, check_ok },
		{ "CVR", 9, check_args },
		{ "USR", 0, client_msn_pre_usr },
		{ NULL, 0, NULL }
	};

CMD	CLIENT_MSNP13_POST_CMDS[] = {
		{ NULL, 0, NULL }
	};

CMD	CLIENT_MSNP13_CMDS[] = {
		{ "FQY", 2, check_args },
		{ "RML", 2, check_args },
		{ "BLP", 2, check_args },
		{ "ADL", 0, msnp13_save_contact },
		{ "CHG", 0, msn_update_status },
		{ "UUX", 2, check_args },
		{ "UUN", 4, check_args },
		{ "URL", 0, msn_url },
		{ "PNG", 0, check_ok },
		{ "QRY", 3, check_args },
		{ "XFR", 2, check_args },
		{ "PRP", 3, check_args },
		{ "OUT", 0, check_ok },
		{ NULL, 0, NULL }
	};

int
check_version(int type, client_ *client, server_ *server) {
 int            version;

    /* check version */
    version = UNKNOWN;
    switch (type) {
	case SERVER_CMD:
	case SERVER_SB_CMD:
	case SERVER_PRE_CMD:
	case SERVER_POST_CMD:
	case SERVER_SB_POST_CMD:
	    if ( server->version & CVR0 ||
		 server->version & MSNP8 ||
		 server->version & MSNP9) {
		version = MSNP8;
            } else if ( server->version & MSNP10 ||
			server->version & MSNP11 ||
			server->version & MSNP12) {
		version = MSNP12;
	    } else if ( server->version & MSNP13 ||
			server->version & MSNP14 ||
			server->version & MSNP15) {
		version = MSNP13;
	    }
	    break;

	case CLIENT_CMD:
	case CLIENT_SB_CMD:
	case CLIENT_PRE_CMD:
	case CLIENT_POST_CMD:
	case CLIENT_SB_POST_CMD:
            if ( client->version & CVR0 ||
		 client->version & MSNP8 ||
		 client->version & MSNP9) {
                version = MSNP8;
	    } else if ( client->version & MSNP10 ||
			client->version & MSNP11 ||
			client->version & MSNP12) {
		version = MSNP12;
	    } else if ( client->version & MSNP13 ||
			client->version & MSNP14 ||
			client->version & MSNP15) {
		version = MSNP13;
	    }
	    break;
    }
    return(version);
}

int
check_sb_version(int type, sb_ *sb) {
 user_		*user = sb->user;
 ns_		*ns   = &user->ns;

    return(check_version(type, ns->client, ns->server));
}

int
check_cmd(user_ *user, command *cmd, int type) {
 unsigned long	erro	= 0;
 client_	*client = user->ns.client;
 server_	*server = user->ns.server;
 log_		*log    = &config.log;
 CMD		*cmds   = (CMD *)0;
 char		*label  = "";
 char		*ep;
 int		version;

    /* check version */
    version = check_version(type, client, server);

    /* assign cmds */
    switch (type) {

	case SERVER_CMD:
	    label = "server_cmd";
	    if (version == MSNP8)
		cmds = SERVER_MSNP8_CMDS;
	    else if (version == MSNP12)
		cmds = SERVER_MSNP12_CMDS;
	    else if (version == MSNP13)
		cmds = SERVER_MSNP13_CMDS;
	    else {
		log->debug("unknown server protocol\n");
		return(RFAIL);
	    }
	    break;

	case SERVER_PRE_CMD:
	    label = "server_pre_cmd";
	    if (version == MSNP8)
		cmds = SERVER_MSNP8_PRE_CMDS;
	    else if (version == MSNP12)
		cmds = SERVER_MSNP12_PRE_CMDS;
	    else if (version == MSNP13)
		cmds = SERVER_MSNP13_PRE_CMDS;
	    else
		cmds = SERVER_PRE_CMDS;
	    break;

	case SERVER_POST_CMD:
	    label = "server_post_cmd";
	    if (version == MSNP8)
		cmds = SERVER_MSNP8_POST_CMDS;
	    else if (version == MSNP12)
		cmds = SERVER_MSNP12_POST_CMDS;
	    else if (version == MSNP13)
		cmds = SERVER_MSNP13_POST_CMDS;
	    break;

	case CLIENT_CMD:
	    label = "client_cmd";
	    if (version == MSNP8)
		cmds = CLIENT_MSNP8_CMDS;
	    else if (version == MSNP12)
		cmds = CLIENT_MSNP12_CMDS;
	    else if (version == MSNP13)
		cmds = CLIENT_MSNP13_CMDS;
	    else {
		log->debug("unknown client protocol\n");
		return(RFAIL);
	    }
	    break;

	case CLIENT_PRE_CMD:
	    label = "client_pre_cmd";
	    if (version == MSNP8)
		cmds = CLIENT_MSNP8_PRE_CMDS;
	    else if (version == MSNP12)
		cmds = CLIENT_MSNP12_PRE_CMDS;
	    else if (version == MSNP13)
		cmds = CLIENT_MSNP13_PRE_CMDS;
	    else
		cmds = CLIENT_PRE_CMDS;
	    break;

	case CLIENT_POST_CMD:
	    label = "client_post_cmd";
	    if (version == MSNP8)
		cmds = CLIENT_MSNP8_POST_CMDS;
	    else if (version == MSNP12)
		cmds = CLIENT_MSNP12_POST_CMDS;
	    else if (version == MSNP13)
		cmds = CLIENT_MSNP13_POST_CMDS;
	    break;
    }

    errno = 0;
    erro = strtoul(cmd->cmd.s, &ep, 10);
    if (errno == 0 && *cmd->cmd.s && (ep && *ep == 0)) {
        log->debug("===> ERRO\n");
        return(ROK);
    }

    for ( ; (cmds && cmds->cmd); cmds++) {
	/* XXX - remove */
	//io_printf(1, "%S: [%S] == [%s]\n", label, cmds->cmd, &cmd->cmd);
	if (strncmp(cmds->cmd, cmd->cmd.s, cmd->cmd.len) == 0) {
	    //print_command(cmd);
	    return(cmds->func(user, cmd, cmds->args));
	}
    }

    /* post commands cannot drop connection */
    switch (type) {
	case SERVER_POST_CMD:
	case CLIENT_POST_CMD:
	    return(ROK);
    }

    log->debug("debug: %S command blocked: ", label);
    print_command(cmd);
    return(RFAIL);
}

SBCMD	SERVER_MSNP8_SB_CMDS[] = {
		{ "USR", 4, check_sb_args },
		{ "CAL", 3, check_sb_args },
		{ "JOI", 0, msn_sb_joi },
		{ "IRO", 0, sb_iro },
		{ "MSG", 0, msn_sb_server_msg },
		{ "ACK", 1, check_sb_args },
		{ "ANS", 2, check_sb_args },
		{ "BYE", 1, check_sb_args },
		{ NULL, 0, NULL }
	};

SBCMD	SERVER_MSNP8_SB_POST_CMDS[] = {
		{ "BYE", 0, msn_sb_post_bye },
		{ NULL, 0, NULL }
	};

SBCMD	CLIENT_MSNP8_SB_CMDS[] = {
		{ "ANS", 4, check_sb_args },
		{ "USR", 3, check_sb_args },
		{ "CAL", 0, msn_sb_cal },
		{ "OUT", 0, check_sb_args },
		{ "MSG", 0, msn_sb_client_msg },
		{ NULL, 0, NULL }
	};

SBCMD   CLIENT_MSNP8_SB_POST_CMDS[] = {
		{ "OUT", 0, msn_sb_post_out },
		{ NULL, 0, NULL }
	};

SBCMD	SERVER_MSNP12_SB_CMDS[] = {
		{ "USR", 4, check_sb_args },
		{ "CAL", 3, check_sb_args },
		{ "IRO", 0, sb_iro },
		{ "JOI", 0, msn_sb_joi },
		{ "ACK", 1, check_sb_args },
		{ "MSG", 0, msn_sb_server_msg },
		{ "BYE", 1, check_sb_args },
		{ "ANS", 2, check_sb_args },
		{ NULL, 0, NULL }
	};

SBCMD	SERVER_MSNP12_SB_POST_CMDS[] = {
		{ "BYE", 0, msn_sb_post_bye },
		{ NULL, 0, NULL }
	};

SBCMD	CLIENT_MSNP12_SB_CMDS[] = {
		{ "USR", 3, check_sb_args },
		{ "CAL", 0, msn_sb_cal },
		{ "ANS", 4, check_sb_args },
		{ "OUT", 0, check_sb_args },
		{ "MSG", 0, msn_sb_client_msg },
		{ NULL, 0, NULL }
	};

SBCMD	CLIENT_MSNP12_SB_POST_CMDS[] = {
		{ "OUT", 0, msn_sb_post_out },
		{ NULL, 0, NULL }
	};

SBCMD	SERVER_MSNP13_SB_CMDS[] = {
		{ "USR", 4, check_sb_args },
		{ "CAL", 3, check_sb_args },
		{ "IRO", 0, sb_iro },
		{ "JOI", 0, msn_sb_joi },
		{ "ACK", 1, check_sb_args },
		{ "MSG", 0, msn_sb_server_msg },
		{ "BYE", 1, check_sb_args },
		{ "ANS", 2, check_sb_args },
		{ NULL, 0, NULL }
	};

SBCMD	SERVER_MSNP13_SB_POST_CMDS[] = {
		{ "BYE", 0, msn_sb_post_bye },
		{ NULL, 0, NULL }
	};

SBCMD	CLIENT_MSNP13_SB_CMDS[] = {
		{ "USR", 3, check_sb_args },
		{ "CAL", 0, msn_sb_cal },
		{ "ANS", 4, check_sb_args },
		{ "OUT", 0, check_sb_args },
		{ "MSG", 0, msn_sb_client_msg },
		{ NULL, 0, NULL }
	};

SBCMD	CLIENT_MSNP13_SB_POST_CMDS[] = {
		{ "OUT", 0, msn_sb_post_out },
		{ NULL, 0, NULL }
	};

int
check_sb_cmd(sb_ *sb, command *cmd, int type) {
 unsigned long	erro   = 0;
 SBCMD		*cmds  = (SBCMD *)0;
 log_		*log   = &config.log;
 char		*label = "";
 char		*ep;
 int		version;

    /* check version */
    version = check_sb_version(type, sb);

    /* assign cmds */
    switch (type) {

	case SERVER_SB_CMD:
	    label = "server_sb_cmd";
	    if (version == MSNP8)
		cmds = SERVER_MSNP8_SB_CMDS;
	    else if (version == MSNP12)
		cmds = SERVER_MSNP12_SB_CMDS;
	    else if (version == MSNP13)
		cmds = SERVER_MSNP13_SB_CMDS;
	    else {
		log->debug("debug: unknown sb server protocol\n");
		return(RFAIL);
	    }
	    break;

	case SERVER_SB_POST_CMD:
	    label = "server_sb_post_cmd";
	    if (version == MSNP8)
		cmds = SERVER_MSNP8_SB_POST_CMDS;
	    else if (version == MSNP12)
		cmds = SERVER_MSNP12_SB_POST_CMDS;
	    else if (version == MSNP13)
		cmds = SERVER_MSNP13_SB_POST_CMDS;
	    else {
		log->debug("debug: unknown sb server protocol\n");
		return(RFAIL);
	    }
	    break;

	case CLIENT_SB_CMD:
	    label = "client_sb_cmd";
	    if (version == MSNP8)
		cmds = CLIENT_MSNP8_SB_CMDS;
	    else if (version == MSNP12)
		cmds = CLIENT_MSNP12_SB_CMDS;
	    else if (version == MSNP13)
		cmds = CLIENT_MSNP13_SB_CMDS;
	    else {
		log->debug("debug: unknown sb client protocol\n");
		return(RFAIL);
	    }
	    break;

	case CLIENT_SB_POST_CMD:
	    label = "client_sb_post_cmd";
	    if (version == MSNP8)
		cmds = CLIENT_MSNP8_SB_POST_CMDS;
	    else if (version == MSNP12)
		cmds = CLIENT_MSNP12_SB_POST_CMDS;
	    else if (version == MSNP13)
		cmds = CLIENT_MSNP13_SB_POST_CMDS;
	    else {
		log->debug("debug: unknown sb client protocol\n");
		return(RFAIL);
	    }
	    break;
    }

    /* check for errors */
    errno = 0;
    erro = strtoul(cmd->cmd.s, &ep, 10);
    if (errno == 0 && *cmd->cmd.s && (ep && *ep == 0)) {
	log->debug("warning: error received\n");
	return(ROK);
    }

    /* dispatch command */
    for ( ; (cmds && cmds->cmd); cmds++) {
	//io_printf(1, "%S: [%S] == [%s]\n", label, cmds->cmd, &cmd->cmd);
	if (strncmp(cmds->cmd, cmd->cmd.s, cmd->cmd.len) == 0) {
	    //print_command(cmd);
	    return(cmds->func(sb, cmd, cmds->args));
	}
    }

    /* post commands cannot drop connection */
    switch (type) {
	case SERVER_SB_POST_CMD:
	case CLIENT_SB_POST_CMD:
	    return(ROK);
    }

    log->debug("debug: sb command blocked: ");
    print_command(cmd);
    return(RFAIL);
}
