/*
 * Copyright (c) 2004, 2005, 2006, 2007 Luiz Otavio O Souza <loos.br@gmail.com>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

static const char rcsid[] = "$Id: contacts.c,v 1.30 2007/09/03 14:18:40 luiz Exp $";

#include <errno.h>
#include <string.h>

#include "io.h"
#include "fmt.h"
#include "sql.h"
#include "xml.h"
#include "return.h"
#include "contacts.h"
#include "protocol.h"
#include "array_cmd.h"
#include "msn-proxy.h"

void contact_zero(contact_ *contact) {
    str_zero(&contact->contact);
    str_zero(&contact->status);
    str_zero(&contact->group);
    str_zero(&contact->uid);
    str_zero(&contact->dn);
    contact->lists = 0;
    contact->flags = 0;
}

void contact_free(contact_ *contact) {
    str_free(&contact->contact);
    str_free(&contact->status);
    str_free(&contact->group);
    str_free(&contact->uid);
    str_free(&contact->dn);
    contact->lists = 0;
    contact->flags = 0;
}

int
msnp13_save_contact(user_ *user, command *cmd, int args) {
 struct xml_tag_opt_	*tag_opt;
 struct xml_tags	xml_tag_head;
 struct xml_tag_	*tag_domain;
 struct xml_tag_	*tag_name;
 struct xml_tag_	*tag;
 contact_		contact;
 string			domain;
 string			name;

    memset(&xml_tag_head, 0, sizeof(xml_tag_head));
    if (xml_parse(&xml_tag_head, &cmd->payload) == RFAIL)
	return(RFAIL);

    tag = SLIST_FIRST(&xml_tag_head);
    if (tag->name.len == 0 || strcmp(tag->name.s, "ml") != 0)
	return(RFAIL);

    SLIST_FOREACH(tag_domain, &tag->xml_tag_head, xml_tag__) {
	if (tag_domain->name.len == 0 ||
		strcmp(tag_domain->name.s, "d") != 0) {
	    continue;
	}
	str_zero(&domain);
	SLIST_FOREACH(tag_opt, &tag_domain->xml_tag_opt_head, xml_tag_opt__) {
	    if (tag_opt->name.len > 0 && strcmp(tag_opt->name.s, "n") == 0) {
		domain.len = tag_opt->value.len;
		domain.s   = tag_opt->value.s;
		break;
	   }
	}

	SLIST_FOREACH(tag_name, &tag_domain->xml_tag_head, xml_tag__) {
	    if (tag_name->name.len == 0 ||
		    strcmp(tag_name->name.s, "c") != 0) {
		return(RFAIL);
	    }

	    contact_zero(&contact);
	    str_zero(&name);
	    SLIST_FOREACH(tag_opt, &tag_name->xml_tag_opt_head, xml_tag_opt__) {
		if (tag_opt->name.len > 0 && strcmp(tag_opt->name.s, "n") == 0) {
		    name.len = tag_opt->value.len;
		    name.s   = tag_opt->value.s;
		} else if (tag_opt->name.len > 0 && strcmp(tag_opt->name.s, "l") == 0) {
		    contact.lists  = atol(tag_opt->value.s);
		    contact.lists |= RL;
		} else if (tag_opt->name.len > 0 && strcmp(tag_opt->name.s, "t") == 0) {
		    /* type */
		}
	    }

	    /* get contact e-mail */
	    contact.contact.len = fmt_printf(NULL, "%s@%s", &name, &domain);
	    if (str_ready(&contact.contact, contact.contact.len + 1) == 0) exit(51);
	    contact.contact.len = fmt_printf(contact.contact.s, "%s@%s",
				  &name, &domain);
	    if (str_copy(&contact.dn, contact.contact.s, contact.contact.len) == 0)
		exit(51);
	    if (str_copy(&contact.status, "OFF", 1) == 0)
		exit(51);
	    if (str_copy(&contact.group, "0", 1) == 0)
		exit(51);
	    if (str_copy(&contact.uid, "0", 1) == 0)
		exit(51);

	    if (sql_save_contact(&user->email, &contact) != ROK) {
		contact_free(&contact);
		return(RFAIL);
	    }

	    contact_free(&contact);
	}
    }

    xml_free(&xml_tag_head);
    return(ROK);
}

int
msnp12_save_contact(user_ *user, command *cmd, int args) {
 unsigned long		tmp;
 contact_		contact;
 string			buf;
 log_			*log = &config.log;
 char			*ep;
 int			i;

    /* zero everything */
    contact_zero(&contact);
    str_zero(&buf);

    /* check min len of cmds */
    if (cmd->args_len < 1)
	return(RFAIL);

    /* get contact email */
    buf.s   = cmd->args[0]->s;
    buf.len = cmd->args[0]->len;

    /* check for min buf len */
    if (buf.len < 2)
	return(RFAIL);

    /* skip "N=" */
    if (strncasecmp(buf.s, "N=", 2) == 0) {
        buf.s   += 2;
        buf.len -= 2;
    }

    /* save contact email */
    contact.contact.len = buf.len;
    contact.contact.s   = buf.s;

    for (i = 1; i < cmd->args_len; i++) {

	str_zero(&buf);
	buf.s   = cmd->args[i]->s;
	buf.len = cmd->args[i]->len;

	if (buf.len >= 2 && strncasecmp(buf.s, "F=", 2) == 0) {
	    /* "F=" == contact display name */
	    buf.len -= 2;
	    buf.s   += 2;
	    if (msn_decode(&buf, &contact.dn) == RFAIL) {
		contact_free(&contact);
		return(RFAIL);
	    }
	} else if (buf.len >= 2 && strncasecmp(buf.s, "C=", 2) == 0) {
	    /* "C=" == contact uid */
	    buf.len -= 2;
	    buf.s   += 2;
	    if (msn_decode(&buf, &contact.uid) == RFAIL) {
		contact_free(&contact);
		return(RFAIL);
	    }
	} else if (buf.len >= 1 && strtoul(buf.s, (char **)0, 10) > 0) {
	    break;
	}
    }

    errno = 0;
    if (i < cmd->args_len) {
	contact.lists = strtoul(cmd->args[i++]->s, &ep, 10);
	if (errno != 0 || (ep && *ep != 0)) {
	    log->debug("debug: fail to decode lists value\n");
	    contact_free(&contact);
	    return(RFAIL);
	}
    }

    /* skip some unknown flag */
    if (i < cmd->args_len) {
	errno = 0;
	tmp = strtoul(cmd->args[i]->s, &ep, 10);
	if (errno == 0 && *cmd->args[i]->s != 0 && ep && *ep == 0) i++;
    }

    /* save contact group */
    if (i < cmd->args_len) {
	if (msn_decode(cmd->args[i], &contact.group) == RFAIL) {
	    contact_free(&contact);
	    return(RFAIL);
	}
    }

    if (sql_save_contact(&user->email, &contact) != ROK) {
	contact_free(&contact);
	return(RFAIL);
    }

    contact_free(&contact);
    return(ROK);
}

int
msnp8_save_contact(user_ *user, command *cmd, int args) {
 contact_		contact;
 char			*ep;

    /* zero everything */
    contact_zero(&contact);

    /* check min len of cmds */
    if (cmd->args_len < 3)
        return(RFAIL);

    /* need to rewrite user syn command ? */
    if (user->save_contacts == NO)
	return(ROK);

    /* save contact email */
    contact.contact.len = cmd->args[0]->len;
    contact.contact.s   = cmd->args[0]->s;

    /* save contact display name */
    if (msn_decode(cmd->args[1], &contact.dn) == RFAIL) {
	contact_free(&contact);
	return(RFAIL);
    }

    /* save contact group */
    if (cmd->args_len == 4) {
	if (msn_decode(cmd->args[3], &contact.group) == RFAIL) {
	    contact_free(&contact);
	    return(RFAIL);
	}
    }

    /* save contact list stats */
    errno = 0;
    contact.lists = strtoul(cmd->args[2]->s, &ep, 10);
    if (errno != 0 || (ep && *ep != 0)) {
	io_printf(1, "ep (%S}\n", ep);
	io_printf(1, "ep (%d}\n", *ep);
	contact_free(&contact);
	return(RFAIL);
    }

    if (sql_save_contact(&user->email, &contact) != ROK) {
	contact_free(&contact);
	return(RFAIL);
    }

    contact_free(&contact);
    return(ROK);
}

int
msn_save_contactgroup(string *email, string *group, string *gid) {
 string		tmp;

    /* zero group */
    str_zero(&tmp);

    /* decode group name */
    if (msn_decode(group, &tmp) == RFAIL)
        return(RFAIL);

    /* save group on database */
    if (sql_save_contactgroup(email, &tmp, gid) != ROK) {
        str_free(&tmp);
        return(RFAIL);
    }

    str_free(&tmp);
    return(ROK);
}

int
msnp12_save_contactgroup(user_ *user, command *cmd, int args) {
    if (user->save_contacts == NO) return(ROK);
    return(msn_save_contactgroup(&user->email, cmd->args[0], cmd->args[1]));
}

int
msnp8_save_contactgroup(user_ *user, command *cmd, int args) {
    if (user->save_contacts == NO) return(ROK);
    return(msn_save_contactgroup(&user->email, cmd->args[1], cmd->args[0]));
}

int
msnp13_contact_initial_state(user_ *user, command *cmd, int args) {
 contact_		contact;
 string			*c;		/* contact */
 string			*s;		/* status */
 string			*d;		/* display name */

    /* zero everything */
    contact_zero(&contact);

    if (cmd->args_len < 7)
	return(RFAIL);

    if ((s = get_arg(cmd, 1)) == NULL ||
		(c = get_arg(cmd, 2)) == NULL ||
		(d = get_arg(cmd, 4)) == NULL) {
	contact_free(&contact);
	return(RFAIL);
    }

    /* check for blocked contacts */
    switch (sql_check_contact_allowed(&user->email, c)) {
	case RFAIL:
	    return(RFAIL);
	case NO:
	    return(RETURN);
	case YES:
	default:
	    break;
    }

    if (d->len > 0 && msn_decode(d, &contact.dn) == RFAIL) {
	contact_free(&contact);
	return(RFAIL);
    }

    if (c->len > 0 && str_copy(&contact.contact, c->s, c->len) == 0) exit(51);
    if (s->len > 0 && str_copy(&contact.status, s->s, s->len) == 0) exit(51);
    contact.flags = atol(cmd->args[5]->s);

    if (sql_set_contact_state(&user->email, &contact) != ROK) {
	contact_free(&contact);
	return(RFAIL);
    }

    contact_free(&contact);

    /* SEEIMG */
    if ((user->commands & SEEIMG) && cmd->args_len == 7) {
	str_free(cmd->args[6]);
	free(cmd->args[6]);
	cmd->args[6] = (string *)0;
	cmd->args_len--;
    }

    return(ROK);
}

int
msn_contact_initial_state(user_ *user, command *cmd, int args) {
 contact_		contact;
 string			*c;		/* contact */
 string			*s;		/* status */
 string			*d;		/* display name */

    /* zero everything */
    contact_zero(&contact);

    if (cmd->args_len < 5)
	return(RFAIL);

    if ((s = get_arg(cmd, 1)) == NULL ||
		(c = get_arg(cmd, 2)) == NULL ||
		(d = get_arg(cmd, 3)) == NULL) {
	contact_free(&contact);
	return(RFAIL);
    }

    /* check for blocked contacts */
    switch (sql_check_contact_allowed(&user->email, c)) {
	case RFAIL:
	    return(RFAIL);
	case NO:
	    return(RETURN);
	case YES:
	default:
	    break;
    }

    if (d->len > 0 && msn_decode(d, &contact.dn) == RFAIL) {
	contact_free(&contact);
	return(RFAIL);
    }

    if (c->len > 0 && str_copy(&contact.contact, c->s, c->len) == 0) exit(51);
    if (s->len > 0 && str_copy(&contact.status, s->s, s->len) == 0) exit(51);
    contact.flags = atol(cmd->args[4]->s);

    if (sql_set_contact_state(&user->email, &contact) != ROK)
	return(RFAIL);

    contact_free(&contact);

    /* SEEIMG */
    if ((user->commands & SEEIMG) && cmd->args_len == 6) {
	str_free(cmd->args[5]);
	free(cmd->args[5]);
	cmd->args[5] = (string *)0;
	cmd->args_len--;
    }

    return(ROK);
}

int
msn_contact_logoff(user_ *user, command *cmd, int args) {
    if (cmd->args_len < 1)
	return(RFAIL);

    if (sql_contact_logoff(&user->email, cmd->args[0]) != ROK)
	return(RFAIL);

    return(ROK);
}

int
msnp13_contact_change(user_ *user, command *cmd, int args) {
 contact_		contact;
 string			*c;
 string			*d;
 string			*s;

    /* zero everything */
    contact_zero(&contact);

    if (cmd->args_len < 4)
	return(RFAIL);

    if ((s = get_arg(cmd, 0)) == NULL ||
		(c = get_arg(cmd, 1)) == NULL ||
		(d = get_arg(cmd, 3)) == NULL) {
	return(RFAIL);
    }

    /* check for blocked contacts */
    switch (sql_check_contact_allowed(&user->email, c)) {
	case RFAIL:
	    return(RFAIL);
	case NO:
	    return(RETURN);
	case YES:
	default:
	    break;
    }

    if (d->len > 0 && msn_decode(d, &contact.dn) == RFAIL) {
	contact_free(&contact);
	return(RFAIL);
    }

    if (c->len > 0 && str_copy(&contact.contact, c->s, c->len) == 0) exit(51);
    if (s->len > 0 && str_copy(&contact.status, s->s, s->len) == 0) exit(51);
    contact.flags = atol(cmd->args[4]->s);

    if (sql_set_contact_state(&user->email, &contact) != ROK) {
	contact_free(&contact);
	return(RFAIL);
    }

    contact_free(&contact);

    /* SEEIMG */
    if ((user->commands & SEEIMG) && cmd->args_len == 7) {
        str_free(cmd->args[6]);
        free(cmd->args[6]);
        cmd->args[6] = (string *)0;
        cmd->args_len--;
    }

    return(ROK);
}

int
msn_contact_change(user_ *user, command *cmd, int args) {
 contact_		contact;
 string			*c;
 string			*d;
 string			*s;

    /* zero everything */
    contact_zero(&contact);

    if (cmd->args_len < 4)
        return(RFAIL);

    if ((s = get_arg(cmd, 0)) == NULL ||
		(c = get_arg(cmd, 1)) == NULL ||
		(d = get_arg(cmd, 2)) == NULL) {
	contact_free(&contact);
	return(RFAIL);
    }

    /* check for blocked contacts */
    switch (sql_check_contact_allowed(&user->email, c)) {
	case RFAIL:
	    return(RFAIL);
	case NO:
	    return(RETURN);
	case YES:
	default:
	    break;
    }

    if (d->len > 0 && msn_decode(d, &contact.dn) == RFAIL) {
	contact_free(&contact);
	return(RFAIL);
    }

    if (c->len > 0 && str_copy(&contact.contact, c->s, c->len) == 0) exit(51);
    if (s->len > 0 && str_copy(&contact.status, s->s, s->len) == 0) exit(51);
    contact.flags = atol(cmd->args[3]->s);

    if (sql_set_contact_state(&user->email, &contact) != ROK)
        return(RFAIL);

    contact_free(&contact);

    /* SEEIMG */
    if ((user->commands & SEEIMG) && cmd->args_len >= 5) {
	str_free(cmd->args[4]);
	free(cmd->args[4]);
	cmd->args[4] = (string *)0;
	cmd->args_len--;
    }

    return(ROK);
}

int
msnp12_client_syn_rewrite(user_ *user, command *cmd, int args) {
 string		*arg1;
 string		*arg2;

    if (cmd->args_len != 3)
	return(RFAIL);

    /* free arg */
    arg1 = cmd->args[1];
    arg2 = cmd->args[2];
    str_free(arg1);
    str_free(arg2);
    free(arg1);
    free(arg2);

    /* alloc new arg */
    arg1 = str_alloc();
    if (arg1 == (string *)0)
	die_nomem();

    arg2 = str_alloc();
    if (arg2 == (string *)0)
	die_nomem();

    /* rewrite arg */
    if (str_copys(arg1, "0") == 0 || str_copys(arg2, "0") == 0)
	die_nomem();

    /* set pointer arg */
    cmd->args[1] = arg1;
    cmd->args[2] = arg2;

    return(ROK);
}

int
msnp8_client_syn_rewrite(user_ *user, command *cmd, int args) {
 client_	*client = user->ns.client;
 string         *arg1;
 char		*ep;

    if (cmd->args_len != 2)
	return(RFAIL);

    if (user->save_contacts == NO)
	return(ROK);

    /* windows messenger bug */
    if (client->version == MSNP8) {

	/* save list version */
	errno = 0;
	user->msnp8_syn_bug = strtoul(cmd->args[1]->s, &ep, 10);
	if (errno != 0 || (ep && *ep != 0))
	    return(RFAIL);

	user->msnp8_syn_bug++;
    }

    /* free arg */
    arg1 = cmd->args[1];
    str_free(arg1);
    free(arg1);

    /* alloc new arg */
    arg1 = str_alloc();
    if (arg1 == (string *)0)
	die_nomem();

    /* rewrite arg */
    if (str_copys(arg1, "0") == 0)
	die_nomem();

    /* set pointer arg */
    cmd->args[1] = arg1;

    return(ROK);
}

int
msnp8_server_syn_rewrite(user_ *user, command *cmd, int args) {
 client_	*client = user->ns.client;
 string         *arg1;

    if (client->version != MSNP8)
	return(ROK);

    /* free arg */
    arg1 = cmd->args[1];
    str_free(arg1);
    free(arg1);

    /* alloc new arg */
    arg1 = str_alloc();
    if (arg1 == (string *)0)
	die_nomem();

    arg1->len = fmt_printf((char *)0, "%d", user->msnp8_syn_bug);
    if (str_ready(arg1, arg1->len + 1) == 0)
	die_nomem();
    arg1->len = fmt_printf(arg1->s, "%d", user->msnp8_syn_bug);

    /* set pointer arg */
    cmd->args[1] = arg1;

    return(ROK);
}
