/*
 * Copyright (c) 2004, 2005, 2006, 2007 Luiz Otavio O Souza <loos.br@gmail.com>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

static const char rcsid[] = "$Id: sql.c,v 1.53 2007/07/28 18:08:01 luiz Exp $";

#include <errno.h>
#include <stdlib.h>
#include <string.h>

#include "fmt.h"
#include "user.h"
#include "mysql.h"
#include "return.h"
#include "contacts.h"
#include "protocol.h"
#include "array_cmd.h"
#include "msn-proxy.h"

int sql_disconnect_all(void) {
 mysql_		*mysql = &config.mysql;
 string		sql;

    str_zero(&sql);

    /* remove stale sb sessions */
    sql.len = fmt_printf((char *)0, "DELETE FROM sb");
    if (str_ready(&sql, sql.len + 1) == 0) die_nomem();
    sql.len = fmt_printf(sql.s, "DELETE FROM sb");

    if (db_query2(mysql, &sql) == RFAIL) {
	str_free(&sql);
	return(RFAIL);
    }

    /* logoff all contacts */
    sql.len = fmt_printf((char *)0, "UPDATE contacts SET contact_status = 'OFF'");
    if (str_ready(&sql, sql.len + 1) == 0) die_nomem();
    sql.len = fmt_printf(sql.s, "UPDATE contacts SET contact_status = 'OFF'");

    if (db_query2(mysql, &sql) == RFAIL) {
	str_free(&sql);
	return(RFAIL);
    }
    str_free(&sql);

    /* logoff everyone */
    sql.len = fmt_printf((char *)0, "UPDATE users SET status = 'OFF'");
    if (str_ready(&sql, sql.len + 1) == 0)
	die_nomem();
    sql.len = fmt_printf(sql.s, "UPDATE users SET status = 'OFF'");

    if (db_query2(mysql, &sql) == RFAIL) {
	str_free(&sql);
	return(RFAIL);
    }
    str_free(&sql);

    return(ROK);
}

int sql_user_disconnect(string *email) {
 mysql_		*mysql = &config.mysql;
 string		sql;
 char		*fmt;

    str_zero(&sql);

    /* logoff user contacts */
    fmt = "UPDATE contacts SET contact_status = 'OFF' WHERE email = '%q'";
    sql.len = fmt_printf((char *)0, fmt, email->s);
    if(str_ready(&sql, sql.len + 1) == 0)
	die_nomem();
    sql.len = fmt_printf(sql.s, fmt, email->s);

    if (db_query2(mysql, &sql) == RFAIL) {
	str_free(&sql);
	return(RFAIL);
    }

    /* logoff user */
    fmt = "UPDATE users SET status = 'OFF' WHERE email = '%q'";
    sql.len = fmt_printf((char *)0, fmt, email->s);
    if(str_ready(&sql, sql.len + 1) == 0)
	die_nomem();
    sql.len = fmt_printf(sql.s, fmt, email->s);

    if (db_query2(mysql, &sql) == RFAIL) {
	str_free(&sql);
	return(RFAIL);
    }

    return(ROK);
}

int
sql_users_update(user_ *user) {
 mysql_		*mysql = &config.mysql;
 string		sql;
 char		*fmt;

    str_zero(&sql);

    /* user exists ? */
    fmt = "UPDATE users SET last_seen = NOW(), last_addr = '%q' "
	  "WHERE email = '%q'";
    sql.len = fmt_printf((char *)0, fmt, user->addr.s, user->email.s);
    if(str_ready(&sql, sql.len + 1) == 0)
	die_nomem();
    sql.len = fmt_printf(sql.s, fmt, user->addr.s, user->email.s);

    if (db_query2(mysql, &sql) == RFAIL) {
	str_free(&sql);
	return(RFAIL);
    }
    str_free(&sql);

    if (db_affected_rows(&mysql->mysql) > 0) {

	/* user exist */
	return(ROK);

    } else if (db_affected_rows(&mysql->mysql) == 0) {

	/* user does not exist */
	fmt =	"INSERT INTO users (email, display_name, last_seen, last_addr, "
		"connect, save_msg, save_contacts, commands) VALUES "
		"('%q', '%q', NOW(), '%q', (SELECT connect FROM defaults), "
		"(SELECT save_msg FROM defaults), "
		"(SELECT save_contacts FROM defaults), "
		"(SELECT commands FROM defaults))";

	sql.len = fmt_printf((char *)0, fmt, user->email.s,
					     user->email.s, user->addr.s);
	if (str_ready(&sql, sql.len + 1) == 0)
	    die_nomem();
	sql.len = fmt_printf(sql.s, fmt, user->email.s,
					 user->email.s, user->addr.s);

	if (db_query2(mysql, &sql) == RFAIL) {
	    str_free(&sql);
	    return(RFAIL);
	}
	str_free(&sql);

	return(ROK);
    }

    /* notreached */
    return(RFAIL);
}

int
sql_set_status(user_ *user) {
 mysql_		*mysql = &config.mysql;
 string		sql;
 char		fmt[] = "UPDATE users SET status = '%q' WHERE email = '%q'";

    str_zero(&sql);
    sql.len = fmt_printf((char *)0, fmt, user->status.s, user->email.s);
    if (str_ready(&sql, sql.len + 1) == 0)
	die_nomem();
    sql.len = fmt_printf(sql.s, fmt, user->status.s, user->email.s);

    if (db_query2(mysql, &sql) == RFAIL) {
	str_free(&sql);
	return(RFAIL);
    }
    str_free(&sql);

    return(ROK);
}

int sql_update_user_dn(user_ *user) {
 mysql_		*mysql = &config.mysql;
 char		fmt[] = "UPDATE users SET display_name = '%q' "
			"WHERE email = '%q'";
 string		sql;

    str_zero(&sql);
    sql.len = fmt_printf((char *)0, fmt, user->dn.s, user->email.s);
    if (str_ready(&sql, sql.len + 1) == 0)
	return(RFAIL);
    sql.len = fmt_printf(sql.s, fmt, user->dn.s, user->email.s);

    if (db_query2(mysql, &sql) == RFAIL) {
	str_free(&sql);
	return(RFAIL);
    }
    str_free(&sql);

    return(ROK);
}

__uint32_t sql_insert_sb(user_ *user) {
 mysql_		*mysql = &config.mysql;
 string		sql;
 char		fmt [] = "INSERT INTO sb (email, since) VALUES ('%q', NOW())";

    str_zero(&sql);
    sql.len = fmt_printf((char *)0, fmt, user->email.s);
    if(str_ready(&sql, sql.len + 1) == 0)
	die_nomem();
    sql.len = fmt_printf(sql.s, fmt, user->email.s);

    if (db_query2(mysql, &sql) == RFAIL) {
	str_free(&sql);
	return(RFAIL);
    }
    str_free(&sql);

    return((__uint32_t)db_last_id(&mysql->mysql));
}

int sql_remove_sb(int id) {
 mysql_		*mysql = &config.mysql;
 string		sql;
 char		fmt[] = "DELETE FROM sb WHERE id = %d";

    str_zero(&sql);
    sql.len = fmt_printf((char *)0, fmt, id);
    if(str_ready(&sql, sql.len + 1) == 0)
	die_nomem();
    sql.len = fmt_printf(sql.s, fmt, id);

    if (db_query2(mysql, &sql) == RFAIL) {
	str_free(&sql);
	return(RFAIL);
    }
    str_free(&sql);

    return(ROK);
}

int sql_log_join(int sb_id, string *email, string *dn) {
 char		fmt[] = "INSERT INTO log (sb_id, `date`, email, display_name, "
			"type) VALUES (%d, NOW(), '%q', '%q', 'join')";
 mysql_		*mysql = &config.mysql;
 string		sql;

    str_zero(&sql);
    sql.len = fmt_printf((char *)0, fmt, sb_id, email->s, dn->s);
    if(str_ready(&sql, sql.len + 1) == 0)
	die_nomem();
    sql.len = fmt_printf(sql.s, fmt, sb_id, email->s, dn->s);

    if (db_query2(mysql, &sql) == RFAIL) {
	str_free(&sql);
	return(RFAIL);
    }
    str_free(&sql);

    return(ROK);
}

int sql_log_stop2(sb_ *sb, string *email, string *contact_email) {
 string		sql;
 mysql_		*mysql = &config.mysql;
 char		fmt[] = "INSERT INTO log (sb_id, `date`, email, display_name, "
			"type) VALUES (%d, NOW(), '%q', "
			"(SELECT contact_dn FROM contacts WHERE email = '%q' "
			"AND contact_email = '%q'), 'stop')";

    str_zero(&sql);
    sql.len = fmt_printf((char *)0, fmt, sb->id, email->s,
						 email->s, contact_email->s);
    if(str_ready(&sql, sql.len + 1) == 0)
	die_nomem();
    sql.len = fmt_printf(sql.s, fmt, sb->id, email->s,
					     email->s, contact_email->s);

    if (db_query2(mysql, &sql) == RFAIL) {
	str_free(&sql);
	return(RFAIL);
    }
    str_free(&sql);

    return(ROK);
}

int sql_log_stop(sb_ *sb, string *email, string *dn) {
 string		sql;
 mysql_		*mysql = &config.mysql;
 char		fmt[] = "INSERT INTO log (sb_id, `date`, email, display_name, "
			"type) VALUES (%d, NOW(), '%q', '%q', 'stop')";

    str_zero(&sql);
    sql.len = fmt_printf((char *)0, fmt, sb->id, email->s, dn->s);
    if(str_ready(&sql, sql.len + 1) == 0)
	die_nomem();
    sql.len = fmt_printf(sql.s, fmt, sb->id, email->s, dn->s);

    if (db_query2(mysql, &sql) == RFAIL) {
	str_free(&sql);
	return(RFAIL);
    }
    str_free(&sql);

    return(ROK);
}

int sql_log_start(int sb_id, string *email, string *dn) {
 char		fmt[] = "INSERT INTO log (sb_id, `date`, email, display_name, "
			"type) VALUES (%d, NOW(), '%q', '%q', 'start')";
 mysql_		*mysql = &config.mysql;
 string		sql;

    str_zero(&sql);
    sql.len = fmt_printf((char *)0, fmt, sb_id, email->s, dn->s);
    if(str_ready(&sql, sql.len + 1) == 0) die_nomem();
    sql.len = fmt_printf(sql.s, fmt, sb_id, email->s, dn->s);

    if (db_query2(mysql, &sql) == RFAIL) {
	str_free(&sql);
	return(RFAIL);
    }
    str_free(&sql);

    return(ROK);
}

int sql_log_msg(int sb_id, string *email, string *dn, string *msg) {
 char		fmt[] = "INSERT INTO log (sb_id, `date`, email, display_name, "
			"type, content) VALUES (%d, NOW(), '%q', '%q', 'msg', "
			"'%q')";
 mysql_		*mysql = &config.mysql;
 string		sql;

    str_zero(&sql);
    sql.len = fmt_printf((char *)0, fmt, sb_id, email->s, dn->s, msg->s);
    if (str_ready(&sql, sql.len + 1) == 0) die_nomem();
    sql.len = fmt_printf(sql.s, fmt, sb_id, email->s, dn->s, msg->s);

    if (db_query2(mysql, &sql) == RFAIL) {
	str_free(&sql);
	return(RFAIL);
    }
    str_free(&sql);

    return(ROK);
}

int sql_read_defaults(void) {
 char		fmt[] = "SELECT connect, save_msg, save_contacts, "
			"commands, internal_host FROM defaults LIMIT 1";
 defaults_	*defaults = &config.defaults;
 mysql_		*mysql = &config.mysql;
 MYSQL_RES	*res;
 MYSQL_ROW	row;
 string		sql;

    str_zero(&sql);
    sql.len = fmt_printf((char *)0, fmt);
    if (str_ready(&sql, sql.len + 1) == 0) die_nomem();
    sql.a = fmt_printf(sql.s, fmt);

    res = db_query(mysql, &sql);
    if (res == (MYSQL_RES *)0) {
	str_free(&sql);
	return(RFAIL);
    }

    str_free(&sql);

    row = db_fetch_row(res);
    if (row == (MYSQL_ROW)0) {
	res = db_free(res);
	return(RFAIL);
    }

    /* connect */
    if (row[0]) {
	if (strcasecmp(row[0], "YES") == 0)
	    defaults->connect = YES;
	else if (strcasecmp(row[0], "NO") == 0)
	    defaults->connect = NO;
    }

    /* save_msg */
    if (row[1]) {
	if (strcasecmp(row[1], "YES") == 0)
	    defaults->save_msg = YES;
	else if (strcasecmp(row[1], "NO") == 0)
	    defaults->save_msg = NO;
    }

    /* save_contacts */
    if (row[2]) {
	if (strcasecmp(row[2], "YES") == 0)
	    defaults->save_contacts = YES;
	else if (strcasecmp(row[2], "NO") == 0)
	    defaults->save_contacts = NO;
    }

    /* commands */
    if (row[3]) {
	defaults->commands = (int)strtol(row[3], (char **)0, 10);
    }

    /* listen_host */
    if (row[4]) {
	if (str_copys(&defaults->internal_host, row[4]) == 0) {
	    res = db_free(res);
	    return(RFAIL);
	}
    }

    res = db_free(res);
    return(ROK);
}

int
read_acl(char *fmt, string *email) {
 mysql_		*mysql = &config.mysql;
 MYSQL_RES	*res;
 MYSQL_ROW	row;
 string		sql;

    str_zero(&sql);
    sql.len = fmt_printf((char *)0, fmt, email->s);
    if (str_ready(&sql, sql.len + 1) == 0)
        die_nomem();
    sql.len = fmt_printf(sql.s, fmt, email->s);

    res = db_query(mysql, &sql);
    str_free(&sql);

    if (res == (MYSQL_RES *)0)
        return(RFAIL);

    row = db_fetch_row(res);
    if (row == (MYSQL_ROW)0 || row[0] == (char *)0) {
        res = db_free(res);
        return(RFAIL);
    }

    if (strcasecmp(row[0], "YES") == 0) {
	res = db_free(res);
	return(YES);
    }
    else if (strcasecmp(row[0], "NO") == 0) {
	res = db_free(res);
	return(NO);
    }
    return(RFAIL);
}

int         
read_commands(user_ *user) {
 mysql_         *mysql = &config.mysql;
 MYSQL_RES      *res;
 MYSQL_ROW      row;
 string         sql;
 char		*ep;
 char		fmt[] = "SELECT commands FROM users WHERE email = '%q'";
            
    str_zero(&sql);
    sql.len = fmt_printf((char *)0, fmt, user->email.s);
    if (str_ready(&sql, sql.len + 1) == 0)
        die_nomem();
    sql.len = fmt_printf(sql.s, fmt, user->email.s);

    res = db_query(mysql, &sql);
    str_free(&sql);

    if (res == (MYSQL_RES *)0)
        return(RFAIL);

    row = db_fetch_row(res);
    if (row == (MYSQL_ROW)0 || row[0] == (char *)0) {
        res = db_free(res);
        return(RFAIL);
    }

    errno = 0;
    user->commands = strtoul(row[0], &ep, 10);
    if (errno != 0 || (ep && *ep != 0)) {
        res = db_free(res);
        return(RFAIL);
    }

    res = db_free(res);

    return(ROK);
}

int
sql_read_user_acl(user_ *user) {

    /* read connect */
    switch (read_acl("SELECT connect FROM users WHERE email = '%q'",
	    &user->email)) {

	case YES:
	    user->connect = YES;
	    break;

	case NO:
	    user->connect = NO;
	    user->save_msg = NO;
	    user->save_contacts = NO;
	    user->commands = DENY_ALL;
	    return(NO);

	case RFAIL:
	default:
	    return(RFAIL);
    }

    /* read save_msg */
    switch (read_acl("SELECT save_msg FROM users WHERE email = '%q'",
	    &user->email)) {

	case YES:
	    user->save_msg = YES;
	    break;

	case NO:
	    user->save_msg = NO;
	    break;
        
	case RFAIL:
	default:
	    return(RFAIL);
    }

    /* read save_contacts */
    switch (read_acl("SELECT save_contacts FROM users WHERE email = '%q'",
	    &user->email)) {

	case YES:
	    user->save_contacts = YES;
	    break;

	case NO:
	    user->save_contacts = NO;
	    break;

	case RFAIL:
	default:
	    return(RFAIL);
    }

    /* read deny commands */
    if (read_commands(user) == RFAIL)
	return(RFAIL);

    return(ROK);
}

int sql_connection_denied(string *email) {
 char		fmt[] = "INSERT INTO log (date, email, type) VALUES "
			"(NOW(), '%q', 'connection denied')";
 mysql_		*mysql = &config.mysql;
 string		sql;

    str_zero(&sql);
    sql.len = fmt_printf((char *)0, fmt, email->s);
    if (str_ready(&sql, sql.len + 1) == 0) die_nomem();
    sql.len = fmt_printf(sql.s, fmt, email->s);

    if (db_query2(mysql, &sql) == RFAIL) {
	str_free(&sql);
	return(RFAIL);
    }

    str_free(&sql);
    return(ROK);
}

int sql_version_denied(string *email) {
 char		fmt[] = "INSERT INTO log (date, email, type) VALUES "
			"(NOW(), '%q', 'client version denied')";
 mysql_		*mysql = &config.mysql;
 string		sql;

    str_zero(&sql);
    sql.len = fmt_printf((char *)0, fmt, email->s);
    if (str_ready(&sql, sql.len + 1) == 0) die_nomem();
    sql.len = fmt_printf(sql.s, fmt, email->s);

    if (db_query2(mysql, &sql) == RFAIL) {
	str_free(&sql);
	return(RFAIL);
    }

    str_free(&sql);
    return(ROK);
}

int sql_password_error(string *email) {
 char		fmt[] = "INSERT INTO log (date, email, type) VALUES "
			"(NOW(), '%q', 'password error')";
 mysql_		*mysql = &config.mysql;
 string		sql;

    str_zero(&sql);
    sql.len = fmt_printf((char *)0, fmt, email->s);
    if (str_ready(&sql, sql.len + 1) == 0) die_nomem();
    sql.len = fmt_printf(sql.s, fmt, email->s);

    if (db_query2(mysql, &sql) == RFAIL) {
        str_free(&sql);
        return(RFAIL);
    }

    str_free(&sql);
    return(ROK);
}

int sql_check_contact(string *email, contact_ *contact) {
 char		fmt[] = "SELECT contact_uid, contact_lists, contact_dn, "
			"contact_group FROM contacts WHERE email = '%q' AND "
			"contact_email = '%q'";
 mysql_		*mysql = &config.mysql;
 MYSQL_RES	*res;
 MYSQL_ROW	row;
 string		sql;

    str_zero(&sql);
    sql.len = fmt_printf((char *)0, fmt, email->s, contact->contact.s);
    if (str_ready(&sql, sql.len + 1) == 0) die_nomem();
    sql.len = fmt_printf(sql.s, fmt, email->s, contact->contact.s);

    res = db_query(mysql, &sql);
    if (res == (MYSQL_RES *)0) {
	str_free(&sql);
	return(RFAIL);
    }

    str_free(&sql);

    if (db_count(res) == 0) {
	/* contact not found */
	res = db_free(res);
	return(NO);
    }

    /* check for update */
    row = db_fetch_row(res);
    if (!row || !row[0] || !row[1]) {
	res = db_free(res);
	return(RFAIL);
    }

    if ((contact->uid.len > 0 && strcasecmp(row[0], contact->uid.s) != 0) ||
		atol(row[1]) != contact->lists ||
		(contact->dn.len > 0 && strcasecmp(row[2], contact->dn.s) != 0) ||
		(contact->group.len > 0 && strcasecmp(row[3], contact->group.s) != 0)) {
	/* contact need to be updated */
	res = db_free(res);
	return(UPDATE);
    }

    res = db_free(res);
    return(YES);
}

int sql_update_contact(string *email, contact_ *contact) {
 char		fmt[] = "UPDATE contacts SET "
			"contact_uid = '%q', "
			"contact_lists = %d, "
			"contact_group = '%q', "
			"contact_dn = '%q' "
			"WHERE email = '%q' AND contact_email = '%q'";
 mysql_		*mysql = &config.mysql;
 string		sql;

    str_zero(&sql);
    sql.len = fmt_printf((char *)0, fmt,
			contact->uid.len > 0 ? contact->uid.s : (unsigned char *)"",
			contact->lists,
			contact->group.len > 0 ? contact->group.s : (unsigned char *)"",
			contact->dn.len > 0 ? contact->dn.s : (unsigned char *)"",
			email->s, contact->contact.s);
    if (str_ready(&sql, sql.len + 1) == 0) die_nomem();
    sql.len = fmt_printf(sql.s, fmt,
			contact->uid.len > 0 ? contact->uid.s : (unsigned char *)"",
			contact->lists,
			contact->group.len > 0 ? contact->group.s : (unsigned char *)"",
			contact->dn.len > 0 ? contact->dn.s : (unsigned char *)"",
			email->s, contact->contact.s);

    if (db_query2(mysql, &sql) == RFAIL) {
	str_free(&sql);
	return(RFAIL);
    }

    str_free(&sql);
    return(ROK);
}

int sql_insert_contact(string *email, contact_ *contact) {
 char		fmt[] = "INSERT INTO contacts (email, contact_uid, "
			"contact_email, contact_lists, contact_dn, "
			"contact_group) VALUES ('%q', '%q', '%q', "
			"%d, '%q', '%q')";
 mysql_		*mysql = &config.mysql;
 string		sql;

    str_zero(&sql);
    sql.len = fmt_printf((char *)0, fmt, email->s,
		contact->uid.len > 0 ? contact->uid.s : (unsigned char *)"",
		contact->contact.s, contact->lists,
		contact->dn.len > 0 ? contact->dn.s : (unsigned char *)"",
		contact->group.len > 0 ? contact->group.s : (unsigned char *)"");
    if (str_ready(&sql, sql.len + 1) == 0) die_nomem();
    sql.len = fmt_printf(sql.s, fmt, email->s,
		contact->uid.len > 0 ? contact->uid.s : (unsigned char *)"",
		contact->contact.s, contact->lists,
		contact->dn.len > 0 ? contact->dn.s : (unsigned char *)"",
		contact->group.len > 0 ? contact->group.s : (unsigned char *)"");

    if (db_query2(mysql, &sql) == RFAIL) {
	str_free(&sql);
	return(RFAIL);
    }

    str_free(&sql);
    return(ROK);
}

int sql_save_contact(string *email, contact_ *contact) {
    switch (sql_check_contact(email, contact)) {

	case UPDATE:
	    /* contact changed - update */
	    return(sql_update_contact(email, contact));

	case NO:
	    /* contact not in list - add */
	    return(sql_insert_contact(email, contact));

	case YES:
	    /* contact in list - do nothing */
	    break;

	case RFAIL:
	default:
	    return(RFAIL);
    }
    return(ROK);
}

int sql_check_contactgroup(string *email, string *name, string *id) {
 char           fmt[] = "SELECT name, id FROM contact_groups WHERE "
			"email = '%q' AND name = '%q'";
 mysql_		*mysql = &config.mysql;
 MYSQL_RES	*res;
 MYSQL_ROW	row;
 string		sql;

    str_zero(&sql);
    sql.len = fmt_printf((char *)0, fmt, email->s, name->s);
    if (str_ready(&sql, sql.len + 1) == 0) die_nomem();
    sql.len = fmt_printf(sql.s, fmt, email->s, name->s);

    res = db_query(mysql, &sql);
    if (!res) {
	str_free(&sql);
	return(RFAIL);
    }

    str_free(&sql);

    if (db_count(res) == 0) {
	/* contact not found */
	res = db_free(res);
	return(NO);
    }

    /* check for update */
    row = db_fetch_row(res);
    if (!row || !row[0] || !row[1]) {
	res = db_free(res);
	return(RFAIL);
    }

    if (strcasecmp(row[0], name->s) != 0 ||
                                strcasecmp(row[1], id->s) != 0) {
	/* contact need to be updated */
	res = db_free(res);
	return(UPDATE);
    }

    res = db_free(res);
    return(YES);
}

int
sql_update_contactgroup(string *email, string *name, string *id) {
 char           fmt[] = "UPDATE contact_groups SET name = '%q', id = '%q' "
			"WHERE email = '%q' AND name = '%q'";
 mysql_         *mysql = &config.mysql;
 string         sql;

    str_zero(&sql);
    sql.len = fmt_printf((char *)0, fmt, name->s, id->s, email->s, name->s);
    if (str_ready(&sql, sql.len + 1) == 0) die_nomem();
    sql.len = fmt_printf(sql.s, fmt, name->s, id->s, email->s, name->s);

    if (db_query2(mysql, &sql) == RFAIL) {
        str_free(&sql);
        return(RFAIL);
    }

    str_free(&sql);
    return(ROK);
}

int
sql_insert_contactgroup(string *email, string *name, string *id) {
 char           fmt[] = "INSERT INTO contact_groups (email, name, id) "
                        "VALUES ('%q', '%q', '%q')";
 mysql_         *mysql = &config.mysql;
 string         sql;

    str_zero(&sql);
    sql.len = fmt_printf((char *)0, fmt, email->s, name->s, id->s);
    if (str_ready(&sql, sql.len + 1) == 0) die_nomem();
    sql.len = fmt_printf(sql.s, fmt, email->s, name->s, id->s);

    if (db_query2(mysql, &sql) == RFAIL) {
        str_free(&sql);
        return(RFAIL);
    }

    str_free(&sql);
    return(ROK);
}

int
sql_save_contactgroup(string *email, string *name, string *id) {
    switch (sql_check_contactgroup(email, name, id)) {

	case UPDATE:
	    /* contact changed - update */
	    return(sql_update_contactgroup(email, name, id));

	case NO:
	    /* contact not in list - add */
	    return(sql_insert_contactgroup(email, name, id));

	case YES:
	    /* contact in list - do nothing */
	    break;

	case RFAIL:
	default:
	    return(RFAIL);
    }
    return(ROK);
}

int
sql_set_contact_state(string *email, contact_ *contact) {
 char           fmt[] = "UPDATE contacts SET "
			"contact_status = '%q', "
			"contact_dn = '%q', "
			"contact_flags = %d "
			"WHERE email = '%q' AND contact_email = '%q'";
 mysql_         *mysql = &config.mysql;
 string         sql;

    str_zero(&sql);
    sql.len = fmt_printf((char *)0, fmt, contact->status.s, contact->dn.s,
			contact->flags,	email->s, contact->contact.s);
    if (str_ready(&sql, sql.len + 1) == 0) die_nomem();
    sql.len = fmt_printf(sql.s, fmt, contact->status.s, contact->dn.s,
			contact->flags,	email->s, contact->contact.s);

    if (db_query2(mysql, &sql) == RFAIL) {
        str_free(&sql);
        return(RFAIL);
    }

    str_free(&sql);
    return(ROK);
}

int
sql_contact_logoff(string *email, string *contact) {
 char		fmt[] = "UPDATE contacts SET "
			"contact_status = 'OFF' "
			"WHERE email = '%q' AND contact_email = '%q'";
 mysql_		*mysql = &config.mysql;
 string		sql;

    str_zero(&sql);
    sql.len = fmt_printf((char *)0, fmt, email->s, contact->s);
    if (str_ready(&sql, sql.len + 1) == 0) die_nomem();
    sql.len = fmt_printf(sql.s, fmt, email->s, contact->s);

    if (db_query2(mysql, &sql) == RFAIL) {
        str_free(&sql);
        return(RFAIL);
    }

    str_free(&sql);
    return(ROK);
}

int
sql_zero_ubx(string *email, string *contact) {
 char		fmt[] = "UPDATE contacts SET "
			"contact_info = '', contact_media = '' "
			"WHERE email = '%q' AND contact_email = '%q'";
 mysql_		*mysql = &config.mysql;
 string		sql;

    str_zero(&sql);
    sql.len = fmt_printf((char *)0, fmt, email->s, contact->s);
    if (str_ready(&sql, sql.len + 1) == 0) die_nomem();
    sql.len = fmt_printf(sql.s, fmt, email->s, contact->s);

    if (db_query2(mysql, &sql) == RFAIL) {
        str_free(&sql);
        return(RFAIL);
    }

    str_free(&sql);
    return(ROK);
}

int
sql_set_ubx_info(string *email, string *contact, string *info) {
 char		fmt[] = "UPDATE contacts SET "
			"contact_info = '%q' "
			"WHERE email = '%q' AND contact_email = '%q'";
 mysql_		*mysql = &config.mysql;
 string		sql;

    str_zero(&sql);
    sql.len = fmt_printf((char *)0, fmt, info->s, email->s, contact->s);
    if (str_ready(&sql, sql.len + 1) == 0) die_nomem();
    sql.len = fmt_printf(sql.s, fmt, info->s, email->s, contact->s);

    if (db_query2(mysql, &sql) == RFAIL) {
	str_free(&sql);
	return(RFAIL);
    }

    str_free(&sql);
    return(ROK);
}

int
sql_check_contact_allowed(string *email, string *contact) {
 char		fmt[] = "SELECT contact_allowed FROM contacts "
			"WHERE email = '%q' AND contact_email = '%q'";
 mysql_		*mysql = &config.mysql;
 string		sql;
 MYSQL_RES	*res;
 MYSQL_ROW	row;

    str_zero(&sql);
    sql.len = fmt_printf((char *)0, fmt, email->s, contact->s);
    if (str_ready(&sql, sql.len + 1) == 0) die_nomem();
    sql.len = fmt_printf(sql.s, fmt, email->s, contact->s);

    res = db_query(mysql, &sql);
    if (res == NULL) {
	str_free(&sql);
	return(RFAIL);
    }

    str_free(&sql);
    row = db_fetch_row(res);
    if (row == NULL) {
	/* no contact */
	res = db_free(res);
	return(YES);
    }

    if (row[0] && strcasecmp(row[0], "NO") == 0) {
	res = db_free(res);
	return(NO);
    }

    res = db_free(res);
    return(YES);
}

int
sql_read_warning(string *warndn, string *warnmsg, string *warnemail,
		 string *msgfont, string *msgcolor) {
 char		fmt[] = "SELECT warndn, warnmsg, warnemail, msgfont, msgcolor "
			"FROM defaults LIMIT 1";
 mysql_		*mysql = &config.mysql;
 string		tmp;
 string		sql;
 MYSQL_RES	*res;
 MYSQL_ROW	row;

    str_zero(&sql);
    sql.len = fmt_printf((char *)0, fmt);
    if (str_ready(&sql, sql.len + 1) == 0) die_nomem();
    sql.len = fmt_printf(sql.s, fmt);

    res = db_query(mysql, &sql);
    if (res == NULL) {
	str_free(&sql);
	return(RFAIL);
    }

    str_free(&sql);
    row = db_fetch_row(res);
    if (row == NULL)
	goto error;

    /* warndn */
    str_zero(&tmp);
    if (row[0]) {
	tmp.s = row[0];
	tmp.len = strlen(row[0]);
	if (msn_encode(&tmp, warndn, 0) == RFAIL) goto error;
    }

    /* warnmsg */
    str_zero(&tmp);
    if (row[1]) {
	tmp.s = row[1];
	tmp.len = strlen(row[1]);
	if (msn_encode(&tmp, warnmsg, 1) == RFAIL) goto error;
    }

    /* warnemail */
    str_zero(&tmp);
    if (row[2]) {
	tmp.s = row[2];
	tmp.len = strlen(row[2]);
	if (msn_encode(&tmp, warnemail, 0) == RFAIL) goto error;
    }

    /* msgfont */
    str_zero(&tmp);
    if (row[3]) {
	tmp.s = row[3];
	tmp.len = strlen(row[3]);
	if (msn_encode(&tmp, msgfont, 0) == RFAIL) goto error;
    }

    /* msgcolor */
    str_zero(&tmp);
    if (row[4]) {
	tmp.s = row[4];
	tmp.len = strlen(row[4]);
	if (msn_encode(&tmp, msgcolor, 0) == RFAIL) goto error;
    }

    res = db_free(res);
    return(ROK);

error:
    res = db_free(res);
    return(RFAIL);
}

