/*
 * Copyright (c) 2004, 2005, 2006, 2007 Luiz Otavio O Souza <loos.br@gmail.com>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

static const char rcsid[] = "$Id: user.c,v 1.26 2007/07/28 18:08:01 luiz Exp $";

#include <string.h>

#include "ns.h"
#include "sb.h"
#include "sql.h"
#include "user.h"
#include "net-io.h"
#include "return.h"
#include "protocol.h"
#include "array_cmd.h"
#include "msn-proxy.h"

unsigned int	user_id  = 0;
unsigned int	user_len = 0;
unsigned int	user_max = 0;

int
user_cmp(struct user__ *user1, struct user__ *user2) {
    return (user1->id - user2->id);
}

RB_HEAD(user_tree_, user__) user_head_ = RB_INITIALIZER(&user_head_);
RB_GENERATE(user_tree_, user__, entry, user_cmp)

int
msnp12_update_dn(user_ *user, command *cmd, int args) {

    /* SYN friendly name */
    if (cmd->args_len == 2 && check_arg(cmd->args[0], "MFN") == ROK) {
	if (msn_decode(cmd->args[1], &user->dn) == RFAIL)
	    return(RFAIL);
    } else if (cmd->args_len == 3 && check_arg(cmd->args[1], "MFN") == ROK) {
	if (msn_decode(cmd->args[2], &user->dn) == RFAIL)
	    return(RFAIL);
    } else
	/* PRP PHN and others */
	return(ROK);

    if (sql_update_user_dn(user) == RFAIL)
	return(RFAIL);

    return(ROK);
}

int
msn_update_status(user_ *user, command *cmd, int args) {
 string		*status = cmd->args[1];
 string		*img    = cmd->args[3];

    /* save user status */
    if (str_copy(&user->status, status->s, status->len) == 0)
	die_nomem();

    /* update sql */
    if (sql_set_status(user) == RFAIL)
	return(RFAIL);

    /* XXX - check flags acl */

    /* check img acl */
    if ( (user->commands & USEIMG) && cmd->args_len == 4) {
	str_free(img);
	free(img);
	cmd->args[3] = (string *)0;
	cmd->args_len--;
    }

    return(ROK);
}

int
msn_url(user_ *user, command *cmd, int args) {
    return(ROK);
}

int
msnp12_ubx(user_ *user, command *cmd, int args) {
 string		*contact;
 string		info;
 string		buf;
 string		tmp;
 int		state;

    /* check for contact */
    if (cmd->args_len < 1)
	return(RFAIL);
    contact = cmd->args[0];

    /* no payload - reset database information */
    if (cmd->payload.len == 0) {
	if (sql_zero_ubx(&user->email, contact) != ROK)
	    return(RFAIL);
	return(ROK);
    }

    state = END;
    str_zero(&buf);
    str_zero(&tmp);
    buf.len = cmd->payload.len;
    buf.s   = cmd->payload.s;
    while (buf.len > 0) {

	/* check for tags */
	if (*buf.s == '<') {
	    if (buf.len >= 5 && strncasecmp("<PSM>", buf.s, 5) == 0) {
		buf.len -= 5;
		buf.s   += 5;
		state = ROK;
		continue;
	    }
	    if (buf.len >= 6 && strncasecmp("</PSM>", buf.s, 6) == 0) {
		buf.len -= 6;
		buf.s   += 6;
		state = END;
		continue;
	    }
	}

	/* copy data */
	if (state == ROK && str_cat(&tmp, buf.s, 1) == 0)
	    return(RFAIL);

	buf.len -= 1;
	buf.s   += 1;
    }

    /* no info - reset database information */
    if (tmp.len == 0) {
	if (sql_zero_ubx(&user->email, contact) != ROK)
	    return(RFAIL);
	return(ROK);
    }

    /* decode info */
    str_zero(&info);
    if (msn_decode(&tmp, &info) == RFAIL) {
	str_free(&tmp);
	return(RFAIL);
    }

    if (sql_set_ubx_info(&user->email, contact, &info) != ROK) {
	str_free(&tmp);
	str_free(&info);
	return(RFAIL);
    }

    str_free(&tmp);
    str_free(&info);

    return(ROK);
}

user_ *
user_free(user_ *user) {
    str_free(&user->dn);
    str_free(&user->addr);
    str_free(&user->email);
    free(user);
    return((user_ *)0);
}

user_ *
user_alloc(struct sockaddr *sa, socklen_t sa_len) {
 user_		*user = (user_ *)0;
 log_		*log = &config.log;
  
    /* alloc new user */
    user = (user_ *)malloc(sizeof(user_));
    if (user == (user_ *)0)
	die_nomem();

    /* zero everything */
    bzero(user, sizeof(user_));

    user->id = ++user_id;
    if (++user_len > user_max)
	user_max = user_len;

    user->state	= PRE_AUTH;
    LIST_INIT(&user->sb_head);

    /* get the user ip address */
    if (resolve_client(&user->addr, sa, sa_len) == RFAIL) {
        log->debug("debug: fail to get ns client address\n");
	user_len--;
	user = user_free(user);
        return((user_ *)0);
    }
 
    if (RB_INSERT(user_tree_, &user_head_, user)) {
	log->debug("debug: fail to insert new user on tree (already there)\n");
	user_len--;
	user = user_free(user);
	return((user_ *)0);
    }

    log->debug("debug: connection from [%s]\n", &user->addr);
    return(user);
}

void
user_disconnect(user_ *user) {
 client_	*proxy = user->ns.xfr_proxy;

    if (user->state == USR_TWN_S)
        sql_password_error(&user->email);

    /* free old proxy xfr address */
    if (proxy != (client_ *)0) {

	/* free xfr_proxy */
	proxy_free(proxy);
	user->ns.xfr_proxy = (client_ *)0;
	free(proxy);
    }

    /* close all SB connections */
    sb_disconnect_all(user);

    /* close connection to ns client */
    user->ns.client = client_disconnect(user->ns.client);

    /* close connection to ns server */
    user->ns.server = server_disconnect(user->ns.server);

    /* remove from sql */
    if (user->email.len > 0)
	sql_user_disconnect(&user->email);

    /* decrease user counter */
    user_len--;

    /* remove user from tree and memory */
    RB_REMOVE(user_tree_, &user_head_, user);
    user = user_free(user);
}

